/*
 * HASH.H - Generic hash handler header file.
 *
 * $Header:   C:/fe/shared/vcs/hash.h_v   1.2   04 Jun 1990 15:17:04   rji  $
 *
 * $Log:   C:/fe/shared/vcs/hash.h_v  $
 * 
 *    Rev 1.2   04 Jun 1990 15:17:04   rji
 * Added hash_delete() cover macro for hdelete().  Expanded HASH_HDR struct
 * to include pointers to randomizing, comparison and copy functions.
 * Added new cover macros to extract these functions.  Changed 
 * hash_create() cover macro and hcreate() function prototype to agree 
 * with new hcreate() function definition.  Changed prototypes of other
 * HASH.C functions to agree with their new definitions.  All needed 
 * #include files now included.  Improved comments.
 * 
 *    Rev 1.1   03 Sep 1989 23:11:26   rji
 * Changed #include file specifiers from <...> to "..." for non-standard
 * headers.
 * 
 *    Rev 1.0   07 Dec 1988 23:10:26   rji
 * Initial revision.
 *   
*/

#if !defined(HASH_H)
#define  HASH_H

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         HASH.H Includes

  $PATHS$
       Functions\H
       Shared\Includes
       Shared\Utils
       Modules\$./FL$

   $3$

   See Also:
        $/SEE(QUEUE.H Macros)$

   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* INCLUDES                                                                */
/*-------------------------------------------------------------------------*/

#include <stddef.h>

#include "queue.h"
/*                                                                   $end$ */

/*-------------------------------------------------------------------------*/
/* a hash table is represented by a pointer to a header structure          */
/* containing control variables for hash table management.                 */
/*-------------------------------------------------------------------------*/

typedef struct hash_hdr
{
   size_t      table_size;    /* # of entries in hash table                */
   size_t      tally;         /* number of elements in hash                */
   size_t      name_size;     /* max significant # of characters in name   */
   int         _i;            /* hash manipulation scratch variable        */
   QUEUE      *table;         /* pointer to hash table                     */
   size_t      (*rand)( struct hash_hdr *, void * );  /* randomizing func  */
   int         (*comp)( void *, void * );          /* name match function  */
   void        (*copy)( void *, void *, size_t );  /* name copy function   */
}
HASH_HDR;

typedef HASH_HDR *HASH;

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         HASH.H Macros

  $PATHS$
       Functions\H
       Shared\Macros
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* MACRO definitions.  All externally accessible hash functions are        */
/* given long names for readability via macros; actual function names      */
/* are kept short for portability.  Hash functions should always be        */
/* invoked via macros.  Macro arguments below are interpreted as follows:  */
/*                                                                         */
/*             ep   = pointer to hash element (as user sees it)            */
/*             es   = element size (size of data to put in table)          */
/*             h    = object of type HASH                                  */
/*             rndf = randomizing ("hash") function for table to use       */
/*             cmpf = name comparison function for table to use            */
/*             cpyf = name copying function for table to use               */
/*             n    = name (hash key)                                      */
/*             ns   = maximum name size (must include space for terminal   */
/*                    null byte if name is string-valued)                  */
/*             t    = element pointer type (i.e., type of ep)              */
/*             ts   = hash table size                                      */
/*                                                                         */
/*-------------------------------------------------------------------------*/

#define  hash_add(h,n,es)        hadd((h),(n),(es))
#define  hash_clear(h)           hclear((h))
#define  hash_count(h)           ((h)->tally)
#define  hash_comp_func(h)       ((h)->comp)
#define  hash_copy_func(h)       ((h)->copy)
#define  hash_create(ts,ns,rndf,cmpf,cpyf)  \
                                 hcreate((ts),(ns),(rndf),(cmpf),(cpyf))
#define  hash_delete(h,ep)       hdelete((h),&(ep))
#define  hash_destroy(h)         hdestroy(&(h))
#define  hash_dump(h)            hdump(h)
#define  hash_find(h,n)          hfind((h),(n))
#define  hash_isempty(h)         (((h)->tally) == 0)
#define  hash_name(h,ep)         ((char *)(ep) - (h)->name_size)
#define  hash_name_size(h)       ((h)->name_size)
#define  hash_rand_func(h)       ((h)->rand)
#define  hash_size(h)            ((h)->table_size)

#define  FOR_EACH_HASH_ENTRY(h,ep,t)                              \
{                                                                 \
   char *_p;                                                      \
   for ( (h)->_i = 0; (h)->_i < (h)->table_size; ++(h)->_i )      \
      if ( (h->table)[(h)->_i] != NULL )                          \
         FOR_EACH_Q_ELE( (h->table)[(h)->_i], _p, char * ) {      \
            (ep) = (t) (_p + (h)->name_size);
#define  END_FOR_HASH   }}
/*                                                                   $end$ */

/*-------------------------------------------------------------------------*/
/* FUNCTION declarations.                                                  */
/*-------------------------------------------------------------------------*/

void    *hadd( HASH, void *, size_t );
void     hclear( HASH );
HASH     hcreate( size_t, size_t,
                  size_t (*)( HASH, void * ),
                  int (*)( void *, void * ),
                  void (*)( void *, void *, size_t ) );
int      hdelete( HASH, void ** );
void     hdestroy( HASH * );
void     hdump( HASH );
void    *hfind( HASH, void * );

#endif /* if !defined(HASH_H) */
